@@ -12,10 +12,12 @@ from spectacles.exceptions import SpectaclesException, LookerApiError
 DEFAULT_API_VERSION = 4.0
 TIMEOUT_SEC = 300
 MAX_ASYNC_CONNECTIONS = 200
+DEFAULT_MAX_TRIES = 3
 BACKOFF_EXCEPTIONS = (
     TimeoutException,
     HTTPStatusError,
     ConnectError,
+    LookerApiError,
 )
 
 
@@ -137,7 +139,6 @@ class LookerClient:
             f"using Looker API {self.api_version}"
         )
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=3)
     async def request(self, method: str, url: str, *args, **kwargs) -> httpx.Response:
         if self.access_token and self.access_token.expired:
             logger.debug("Looker API access token has expired, requesting a new one")
@@ -161,6 +162,7 @@ class LookerClient:
     async def delete(self, url, *args, **kwargs) -> httpx.Response:
         return await self.request("DELETE", url, *args, **kwargs)
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     def get_looker_release_version(self) -> str:
         """Gets the version number of connected Looker instance.
 
@@ -191,6 +193,7 @@ class LookerClient:
 
         return response.json()["looker_release_version"]
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_workspace(self) -> str:
         """Gets the session workspace.
 
@@ -218,6 +221,7 @@ class LookerClient:
             ) from error
         return response.json()["workspace_id"]
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def update_workspace(self, workspace: str) -> None:
         """Updates the session workspace.
 
@@ -245,6 +249,7 @@ class LookerClient:
             ) from error
         self.workspace = workspace
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_all_branches(self, project: str) -> List[str]:
         """Returns a list of git branches in the project repository.
 
@@ -272,6 +277,7 @@ class LookerClient:
 
         return [branch["name"] for branch in response.json()]
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def checkout_branch(self, project: str, branch: str) -> None:
         """Checks out a new git branch. Only works in dev workspace.
 
@@ -298,6 +304,7 @@ class LookerClient:
                 response=response,
             ) from error
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def reset_to_remote(self, project: str) -> None:
         """Reset a project development branch to the revision of the project that is on the remote.
 
@@ -324,6 +331,7 @@ class LookerClient:
                 response=response,
             ) from error
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_manifest(self, project: str) -> JsonDict:
         """Gets all the dependent LookML projects defined in the manifest file.
 
@@ -356,6 +364,7 @@ class LookerClient:
 
         return manifest
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_active_branch(self, project: str) -> JsonDict:
         """Gets the active branch for the user in the given project.
 
@@ -389,11 +398,13 @@ class LookerClient:
 
         return response.json()
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_active_branch_name(self, project: str) -> str:
         """Helper method to return only the branch name."""
         full_response = await self.get_active_branch(project)
         return full_response["name"]
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def create_branch(
         self, project: str, branch: str, ref: Optional[str] = None
     ) -> None:
@@ -437,6 +448,7 @@ class LookerClient:
                 response=response,
             ) from error
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def hard_reset_branch(self, project: str, branch: str, ref: str) -> None:
         """Hard resets a branch to the ref prodvided.
 
@@ -471,6 +483,7 @@ class LookerClient:
                 response=response,
             ) from error
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def delete_branch(self, project: str, branch: str):
         """Deletes a branch in the given project.
 
@@ -499,6 +512,7 @@ class LookerClient:
                 response=response,
             ) from error
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def all_lookml_tests(self, project: str) -> List[JsonDict]:
         """Gets all LookML/data tests for a given project.
 
@@ -531,6 +545,7 @@ class LookerClient:
 
         return response.json()
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def run_lookml_test(
         self, project: str, model: Optional[str] = None, test: Optional[str] = None
     ) -> List[JsonDict]:
@@ -587,6 +602,7 @@ class LookerClient:
 
         return response.json()
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_lookml_models(self, fields: Optional[List] = None) -> List[JsonDict]:
         """Gets all models and explores from the LookmlModel endpoint.
 
@@ -617,6 +633,7 @@ class LookerClient:
 
         return response.json()
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_lookml_dimensions(self, model: str, explore: str) -> List[str]:
         """Gets all dimensions for an explore from the LookmlModel endpoint.
 
@@ -653,7 +670,7 @@ class LookerClient:
 
         return response.json()["fields"]["dimensions"]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=8)
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=5)
     async def create_query(
         self,
         model: str,
@@ -719,7 +736,7 @@ class LookerClient:
         )
         return result
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=8)
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=5)
     async def create_query_task(self, query_id: str) -> str:
         """Runs a previously created query asynchronously and returns the query task ID.
 
@@ -763,6 +780,7 @@ class LookerClient:
         logger.debug("Query %s is running under query task %s", query_id, query_task_id)
         return query_task_id
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def get_query_task_multi_results(
         self, query_task_ids: Tuple[str, ...]
     ) -> JsonDict:
@@ -806,6 +824,7 @@ class LookerClient:
         result = response.json()
         return result
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def cancel_query_task(self, query_task_id: str) -> None:
         """Cancels a query task.
 
@@ -841,6 +860,7 @@ class LookerClient:
         result = response.json()
         return result
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def lookml_validation(self, project) -> JsonDict:
         logger.debug(f"Validating LookML for project '{project}'")
         url = utils.compose_url(self.api_url, path=["projects", project, "validate"])
@@ -860,6 +880,7 @@ class LookerClient:
         result = response.json()
         return result
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def cached_lookml_validation(self, project) -> Optional[JsonDict]:
         logger.debug(f"Getting cached LookML validation results for '{project}'")
         url = utils.compose_url(self.api_url, path=["projects", project, "validate"])
@@ -886,6 +907,7 @@ class LookerClient:
         result = response.json()
         return result
 
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def all_folders(self) -> List[JsonDict]:
         logger.debug("Getting information about all folders")
         url = utils.compose_url(self.api_url, path=["folders"])
@@ -905,7 +927,7 @@ class LookerClient:
         result = response.json()
         return result
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=5)
+    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
     async def run_query(self, query_id: str) -> str:
         """Returns the compiled SQL for a given query ID.
 
