@@ -1,6 +1,7 @@
+import asyncio
 import re
 from spectacles.exceptions import LookerApiError, SqlError
-from typing import Dict, List, Optional, cast, Tuple
+from typing import Dict, List, Optional, Tuple, Set
 from dataclasses import dataclass
 import itertools
 from spectacles.client import LookerClient
@@ -11,10 +12,14 @@ from spectacles.validators import (
     LookMLValidator,
 )
 from spectacles.types import JsonDict
-from spectacles.validators.sql import SqlTest, DEFAULT_CHUNK_SIZE
+from spectacles.validators.sql import (
+    DEFAULT_CHUNK_SIZE,
+    DEFAULT_QUERY_CONCURRENCY,
+    DEFAULT_RUNTIME_THRESHOLD,
+)
 from spectacles.exceptions import SpectaclesException
 from spectacles.utils import time_hash
-from spectacles.lookml import build_project, Project, Explore
+from spectacles.lookml import CompiledSql, build_project, Explore
 from spectacles.logger import GLOBAL_LOGGER as logger
 from spectacles.printer import print_header
 
@@ -38,22 +43,22 @@ class LookerBranchManager:
         client: LookerClient,
         project: str,
         remote_reset: bool = False,
-        pin_imports: Dict[str, str] = {},
+        pin_imports: Optional[Dict[str, str]] = None,
+        skip_imports: Optional[List[str]] = None,
     ):
         """Context manager for Git branch checkout, creation, and deletion."""
         logger.debug(f"Setting up branch manager in project '{project}'")
         self.client = client
         self.project = project
         self.remote_reset = remote_reset
-        self.pin_imports = pin_imports
+        self.pin_imports = pin_imports or {}
+        self.history: List[ProjectState] = []
 
-        state: ProjectState = self.get_project_state()
-        self.workspace: str = state.workspace
-        self.history: List[ProjectState] = [state]
         self.commit: Optional[str] = None
         self.branch: Optional[str] = None
         self.is_temp_branch: bool = False
         self.import_managers: List[LookerBranchManager] = []
+        self.skip_imports: List[str] = [] if skip_imports is None else skip_imports
 
     def __call__(self, ref: Optional[str] = None, ephemeral: Optional[bool] = None):
         logger.debug("")
@@ -90,32 +95,36 @@ class LookerBranchManager:
         self.import_managers = []
         return self
 
-    def __enter__(self):
+    async def __aenter__(self):
         logger.indent(1)
+
+        state: ProjectState = await self.get_project_state()
+        self.workspace: str = state.workspace
+        self.history = [state]
         # A branch was passed, so we check it out in dev mode.
         if self.branch:
-            self.update_workspace("dev")
+            await self.update_workspace("dev")
             if self.ephemeral:
-                self.branch = self.checkout_temp_branch("origin/" + self.branch)
+                self.branch = await self.checkout_temp_branch("origin/" + self.branch)
             else:
-                self.client.checkout_branch(self.project, self.branch)
+                await self.client.checkout_branch(self.project, self.branch)
                 if self.remote_reset:
-                    self.client.reset_to_remote(self.project)
+                    await self.client.reset_to_remote(self.project)
         # A commit was passed, so we non-destructively create a temporary branch we can
         # hard reset to the commit.
         elif self.commit:
-            self.branch = self.checkout_temp_branch(self.commit)
+            self.branch = await self.checkout_temp_branch(self.commit)
         # Neither branch nor commit were passed, so go to production.
         else:
             if self.init_state.workspace == "production":
                 prod_state = self.init_state
             else:
-                self.update_workspace("production")
-                prod_state = self.get_project_state()
+                await self.update_workspace("production")
+                prod_state = await self.get_project_state()
             self.branch = prod_state.branch
             self.commit = prod_state.commit
             if self.ephemeral:
-                self.branch = self.checkout_temp_branch(prod_state.commit)
+                self.branch = await self.checkout_temp_branch(prod_state.commit)
 
         logger.debug(
             f"Set project '{self.project}' to branch '{self.branch}' @ "
@@ -123,7 +132,7 @@ class LookerBranchManager:
             f"[ephemeral = {self.ephemeral}]"
         )
 
-        self.imports: List[str] = self.get_project_imports()
+        self.imports: List[str] = await self.get_project_imports()
         logger.debug(
             f"Project '{self.project}' imports the following projects: {self.imports}"
         )
@@ -137,15 +146,24 @@ class LookerBranchManager:
             )
         else:
             logger.debug("Creating temporary branches in imported projects")
-            previously_imported = tuple(mgr.project for mgr in self.import_managers)
             for project in self.imports:
-                if project not in previously_imported:
+                if project == self.project:
+                    raise SpectaclesException(
+                        name="/errors/circular-project-import",
+                        title="Circular project import",
+                        detail=f"Project '{self.project}' imports itself",
+                    )
+                elif project not in self.skip_imports:
                     import_ref = self.pin_imports.get(project, None)
                     manager = LookerBranchManager(
-                        self.client, project, pin_imports=self.pin_imports
+                        self.client,
+                        project,
+                        pin_imports=self.pin_imports,
+                        skip_imports=self.skip_imports,
                     )
-                    manager(ref=import_ref, ephemeral=True).__enter__()
+                    await manager(ref=import_ref, ephemeral=True).__aenter__()
                     self.import_managers.append(manager)
+                    self.skip_imports.append(project)
                 else:
                     logger.debug(
                         f"Skipping project '{project}', which is already imported"
@@ -154,7 +172,7 @@ class LookerBranchManager:
         logger.indent(-1)
         logger.debug("")
 
-    def __exit__(self, *args):
+    async def __aexit__(self, *args):
         logger.debug("")
         logger.debug(f"Cleaning up Git state in '{self.project}'")
         logger.indent(1)
@@ -166,25 +184,37 @@ class LookerBranchManager:
         logger.debug(message)
 
         if self.is_temp_branch:
+            if self.branch is None:
+                raise TypeError(
+                    "Unable to clean up temporary branch, LookerBranchManager.branch "
+                    "is None"
+                )
             dev_state = self.history.pop()
-            self.client.checkout_branch(self.project, dev_state.branch)
-            self.client.delete_branch(self.project, self.branch)
+            await self.client.checkout_branch(self.project, dev_state.branch)
+            await self.client.delete_branch(self.project, self.branch)
 
         for manager in self.import_managers:
-            manager.__exit__()
+            await manager.__aexit__()
 
         if self.init_state.workspace == "production":
-            self.update_workspace("production")
+            await self.update_workspace("production")
         else:
-            self.update_workspace("dev")
-            self.client.checkout_branch(self.project, self.init_state.branch)
+            await self.update_workspace("dev")
+            await self.client.checkout_branch(self.project, self.init_state.branch)
 
         logger.indent(-1)
         logger.debug("")
 
     @property
     def init_state(self) -> ProjectState:
-        return self.history[0]
+        try:
+            state = self.history[0]
+        except IndexError as error:
+            raise IndexError(
+                "No history exists, you must enter the context manager "
+                "to generate initial state."
+            ) from error
+        return state
 
     @property
     def ref(self) -> Optional[str]:
@@ -193,42 +223,42 @@ class LookerBranchManager:
         else:
             return self.branch
 
-    def update_workspace(self, workspace: str):
+    async def update_workspace(self, workspace: str):
         if workspace not in ("dev", "production"):
             raise ValueError("Workspace can only be set to 'dev' or 'production'")
         if self.workspace != workspace:
-            self.client.update_workspace(workspace)
+            await self.client.update_workspace(workspace)
             self.workspace = workspace
 
-    def get_project_state(self) -> ProjectState:
-        workspace = self.client.get_workspace()
-        branch_info = self.client.get_active_branch(self.project)
+    async def get_project_state(self) -> ProjectState:
+        workspace = await self.client.get_workspace()
+        branch_info = await self.client.get_active_branch(self.project)
         return ProjectState(
             self.project, workspace, branch_info["name"], branch_info["ref"]
         )
 
-    def get_project_imports(self) -> List[str]:
+    async def get_project_imports(self) -> List[str]:
         try:
-            manifest = self.client.get_manifest(self.project)
+            manifest = await self.client.get_manifest(self.project)
         except LookerApiError:
             return []
         else:
             return [p["name"] for p in manifest["imports"] if not p["is_remote"]]
 
-    def checkout_temp_branch(self, ref: str) -> str:
+    async def checkout_temp_branch(self, ref: str) -> str:
         """Creates a temporary branch off a commit or off production."""
         # Save the dev mode state so we have somewhere to delete the temp branch
         # from later. We can't delete branches from prod workspace.
-        self.update_workspace("dev")
-        self.history.append(self.get_project_state())
+        await self.update_workspace("dev")
+        self.history.append(await self.get_project_state())
         name = "tmp_spectacles_" + time_hash()
         logger.debug(
             f"Branching '{name}' off '{ref}'. "
             f"Afterwards, restoring to branch '{self.init_state.branch}' in "
             f"project '{self.project}'"
         )
-        self.client.create_branch(self.project, name)
-        self.client.hard_reset_branch(self.project, name, ref)
+        await self.client.create_branch(self.project, name)
+        await self.client.hard_reset_branch(self.project, name, ref)
         self.is_temp_branch = True
         return name
 
@@ -255,52 +285,55 @@ class Runner:
         client: LookerClient,
         project: str,
         remote_reset: bool = False,
-        pin_imports: Dict[str, str] = {},
+        pin_imports: Optional[Dict[str, str]] = None,
     ):
         self.project = project
         self.client = client
         self.branch_manager = LookerBranchManager(
-            client, project, remote_reset, pin_imports
+            client, project, remote_reset, pin_imports or {}
         )
 
-    def validate_sql(
+    async def validate_sql(
         self,
         ref: Optional[str] = None,
-        filters: List[str] = None,
+        filters: Optional[List[str]] = None,
         fail_fast: bool = True,
         incremental: bool = False,
         target: Optional[str] = None,
-        concurrency: int = 10,
+        concurrency: int = DEFAULT_QUERY_CONCURRENCY,
         profile: bool = False,
-        runtime_threshold: int = 5,
+        runtime_threshold: int = DEFAULT_RUNTIME_THRESHOLD,
         chunk_size: int = DEFAULT_CHUNK_SIZE,
         ignore_hidden_fields: bool = False,
     ) -> JsonDict:
         if filters is None:
             filters = ["*/*"]
         validator = SqlValidator(self.client, concurrency, runtime_threshold)
-        tests: List[SqlTest] = []
-
         ephemeral = True if incremental else None
         # Create explore-level tests for the desired ref
-        with self.branch_manager(ref=ref, ephemeral=ephemeral):
+        async with self.branch_manager(ref=ref, ephemeral=ephemeral):
             base_ref = self.branch_manager.ref  # Resolve the full ref after checkout
-            logger.debug("Building explore tests for the desired ref")
-            project = build_project(
+            logger.debug("Compiling SQL for Explores at the base ref")
+            project = await build_project(
                 self.client,
                 name=self.project,
                 filters=filters,
                 include_dimensions=True,
                 ignore_hidden_fields=ignore_hidden_fields,
             )
-            base_tests = validator.create_tests(
-                project, compile_sql=incremental, chunk_size=chunk_size
-            )
+            base_explores: Set[CompiledSql] = set()
+            if incremental:
+                compiled_explores = await asyncio.gather(
+                    *(
+                        validator.compile_explore(explore)
+                        for explore in project.iter_explores()
+                    )
+                )
+                base_explores = set(compiled_explores)
 
         if incremental:
-            unique_base_tests = set(base_tests)
             # Create explore tests for the target
-            with self.branch_manager(ref=target, ephemeral=True):
+            async with self.branch_manager(ref=target, ephemeral=True):
                 target_ref = self.branch_manager.ref
                 if target_ref == base_ref:
                     raise SpectaclesException(
@@ -313,136 +346,147 @@ class Runner:
                             "so incremental comparison isn't possible."
                         ),
                     )
-                logger.debug("Building explore tests for the target ref")
-                target_project: Project = build_project(
+                logger.debug("Compiling SQL for Explores at the target ref")
+                target_project = await build_project(
                     self.client,
                     name=self.project,
                     filters=filters,
                     include_dimensions=True,
                     ignore_hidden_fields=ignore_hidden_fields,
                 )
-                target_tests = validator.create_tests(
-                    target_project, compile_sql=True, chunk_size=chunk_size
-                )
-                unique_target_tests = set(target_tests)
+                target_explores: Set[CompiledSql] = set()
+                if incremental:
+                    compiled_explores = await asyncio.gather(
+                        *(
+                            validator.compile_explore(explore)
+                            for explore in target_project.iter_explores()
+                        )
+                    )
+                    target_explores = set(compiled_explores)
 
             # Determine which explore tests are identical between target and base
             # Iterate instead of set operations so we have control of which test, and
             # corresponding which `lookml_ref` is used
-            tests = []
-            for test in unique_base_tests:
-                if test in unique_target_tests:
+            explores: Tuple[Explore, ...] = tuple()
+            for compiled in base_explores:
+                explore = project.get_explore(
+                    compiled.model_name, compiled.explore_name
+                )
+                if explore is None:
+                    raise TypeError(
+                        "Couldn't find the explore "
+                        f"{compiled.model_name}.{compiled.explore_name}"
+                    )
+                if compiled in target_explores:
                     # Mark explores with the same compiled SQL (test) as skipped
-                    explore = cast(Explore, test.lookml_ref)  # Appease mypy
                     explore.skipped = True
                 else:
                     # Test explores with unique SQL for base ref
-                    tests.append(test)
+                    explores += (explore,)
 
-            logger.debug(
-                f"Found {len(unique_base_tests - unique_target_tests)} "
-                "explore tests with unique SQL"
-            )
+            logger.debug(f"Found {len(explores)} explores with unique SQL")
         else:
-            tests = base_tests
+            explores = tuple(project.iter_explores())
 
-        explore_count = project.count_explores()
+        explore_count = len(explores)
         print_header(
             f"Testing {explore_count} "
             f"{'explore' if explore_count == 1 else 'explores'} "
             + ("[fail fast] " if fail_fast else "")
-            + f"[concurrency = {validator.query_slots}]"
+            + f"[concurrency = {concurrency}]"
         )
 
-        with self.branch_manager(ref=ref):
-            validator.run_tests(tests, profile=profile if fail_fast else False)
+        async with self.branch_manager(ref=ref):
+            await validator.search(
+                explores, fail_fast, chunk_size, profile=profile if fail_fast else False
+            )
 
         # Create dimension tests for the desired ref when explores errored
-        if not fail_fast:
-            with self.branch_manager(ref=ref, ephemeral=ephemeral):
-                base_ref = self.branch_manager.ref
-                logger.debug("Building dimension tests for the desired ref")
-                base_tests = validator.create_tests(project, at_dimension_level=True)
-                validator.run_tests(base_tests, profile)
-
+        if not fail_fast and incremental:
+            errored_dimensions = project.iter_dimensions(errored=True)
             # For errored dimensions, create dimension tests for the target ref
-            if incremental:
-                with self.branch_manager(ref=target, ephemeral=True):
-                    target_ref = self.branch_manager.ref
-                    logger.debug("Building dimension tests for the target ref")
-
-                    target_sql: List[Tuple[str, str]] = []
-                    for dimension in project.iter_dimensions(errored=True):
-                        test = validator._create_dimension_test(
-                            dimension, compile_sql=True
-                        )
-                        if test.sql:
-                            target_sql.append((test.lookml_ref.name, test.sql))
-
-                # Keep only the errors that don't exist on the target branch
-                logger.debug(
-                    "Removing errors that would exist in project "
-                    f"@ {target or 'production'}"
+            async with self.branch_manager(ref=target, ephemeral=True):
+                target_ref = self.branch_manager.ref
+                logger.debug("Compiling SQL for dimensions at the target ref")
+                compiled_dimensions = await asyncio.gather(
+                    *(
+                        validator.compile_dimension(dimension)
+                        for dimension in project.iter_dimensions(errored=True)
+                    )
                 )
+                target_dimensions = set(compiled_dimensions)
 
-                for dimension in project.iter_dimensions(errored=True):
-                    for error in dimension.errors:
-                        if (
-                            isinstance(error, SqlError)
-                            and (dimension.name, error.metadata["sql"]) in target_sql
-                        ):
-                            error.ignore = True
+            # Keep only the errors that don't exist on the target branch
+            logger.debug(
+                "Removing errors that would exist in project "
+                f"@ {target or 'production'}"
+            )
+
+            # Namespace SQL with the dimension name, just in case
+            target_sql = tuple(
+                (compiled.dimension_name, compiled.sql)
+                for compiled in target_dimensions
+            )
+            for dimension in errored_dimensions:
+                for error in dimension.errors:
+                    if (
+                        isinstance(error, SqlError)
+                        and (dimension.name, error.metadata["sql"]) in target_sql
+                    ):
+                        error.ignore = True
 
         results = project.get_results(validator="sql", fail_fast=fail_fast)
         return results
 
-    def validate_data_tests(
+    async def validate_data_tests(
         self,
         ref: Optional[str] = None,
-        filters: List[str] = None,
+        filters: Optional[List[str]] = None,
     ) -> JsonDict:
         if filters is None:
             filters = ["*/*"]
-        with self.branch_manager(ref):
+        async with self.branch_manager(ref):
             validator = DataTestValidator(self.client)
             logger.info(
                 "Building LookML project hierarchy for project "
                 f"'{self.project}' @ {self.branch_manager.ref}"
             )
-            project = build_project(self.client, name=self.project, filters=filters)
+            project = await build_project(
+                self.client, name=self.project, filters=filters
+            )
             explore_count = project.count_explores()
             print_header(
                 f"Running data tests based on {explore_count} "
                 f"{'explore' if explore_count == 1 else 'explores'}"
             )
-            tests = validator.get_tests(project)
-            validator.validate(tests)
+            tests = await validator.get_tests(project)
+            await validator.validate(tests)
 
         results = project.get_results(validator="data_test")
         return results
 
-    def validate_lookml(self, ref: Optional[str], severity: str) -> JsonDict:
-        with self.branch_manager(ref=ref):
+    async def validate_lookml(self, ref: Optional[str], severity: str) -> JsonDict:
+        async with self.branch_manager(ref=ref):
             validator = LookMLValidator(self.client)
             print_header(f"Validating LookML in project {self.project} [{severity}]")
-            results = validator.validate(self.project, severity)
+            results = await validator.validate(self.project, severity)
         return results
 
-    def validate_content(
+    async def validate_content(
         self,
         ref: Optional[str] = None,
-        filters: List[str] = None,
+        filters: Optional[List[str]] = None,
         incremental: bool = False,
         target: Optional[str] = None,
         exclude_personal: bool = False,
-        folders: List[str] = None,
+        folders: Optional[List[str]] = None,
     ) -> JsonDict:
         if filters is None:
             filters = ["*/*"]
         if folders is None:
             folders = []
 
-        with self.branch_manager(ref=ref):
+        async with self.branch_manager(ref=ref):
             validator = ContentValidator(
                 self.client,
                 exclude_personal,
@@ -452,7 +496,7 @@ class Runner:
                 "Building LookML project hierarchy for project "
                 f"'{self.project}' @ {self.branch_manager.ref}"
             )
-            project = build_project(
+            project = await build_project(
                 self.client,
                 name=self.project,
                 filters=filters,
@@ -464,23 +508,23 @@ class Runner:
                 f"{'explore' if explore_count == 1 else 'explores'}"
                 + (" [incremental mode] " if incremental else "")
             )
-            validator.validate(project)
+            await validator.validate(project)
             results = project.get_results(validator="content", filters=filters)
 
         if incremental and (self.branch_manager.branch or self.branch_manager.commit):
             logger.debug("Starting another content validation against the target ref")
-            with self.branch_manager(ref=target):
+            async with self.branch_manager(ref=target):
                 logger.debug(
                     "Building LookML project hierarchy for project "
                     f"'{self.project}' @ {self.branch_manager.ref}"
                 )
-                target_project = build_project(
+                target_project = await build_project(
                     self.client,
                     name=self.project,
                     filters=filters,
                     include_all_explores=True,
                 )
-                validator.validate(target_project)
+                await validator.validate(target_project)
                 target_results = target_project.get_results(
                     validator="content", filters=filters
                 )
